import requests
import pickle
from copy import deepcopy
import os
import time
import warnings


class UnpywallCache:
    """
    This class stores query results from Unpaywall.
    It has a configurable timeout that can also be set to never expire.

    Attributes
    ----------
    name : string
        The filename used to save and load the cache by default.
    content : dict
        A dictionary mapping dois to requests.Response objects.
    access_times : dict
        A dictionary mapping dois to the datetime when each was last updated.

    """

    def __init__(self, name: str = None, timeout=None) -> None:
        """
        Create a cache object.

        Parameters
        ----------
        timeout : float or int
            The number of seconds that each entry should last.
        name : str
            The filename used to save and load the cache by default.
        """
        if not name:
            self.name = os.path.join(os.getcwd(), 'unpaywall_cache')
        else:
            self.name = name
        try:
            self.load(self.name)
        except FileNotFoundError:
            # print('No cache found. A new cache was initialized.')
            self.reset_cache()
        self.timeout = timeout

    def reset_cache(self) -> None:
        """
        Set the cache to a blank state.
        """
        self.content = {}
        self.access_times = {}
        self.save()

    def delete(self, doi: str) -> None:
        """
        Remove an individual doi from the cache.

        Parameters
        ----------
        doi : str
            The DOI to be removed from the cache.
        """
        if doi in self.access_times:
            del self.access_times[doi]
        if doi in self.content:
            del self.content[doi]
        self.save()

    def timed_out(self, doi: str) -> bool:
        """
        Return whether the record for the given doi has expired.

        Parameters
        ----------
        doi : str
            The DOI to be removed from the cache.

        Returns
        -------
        is_timed_out : bool
            Whether the given entry has timed out.
        """
        if not self.timeout:
            is_timed_out = False
        else:
            is_timed_out = time.time() > self.access_times[doi] + self.timeout
        return is_timed_out

    def get(self, doi: str, errors: str = 'raise',
            force: bool = False, ignore_cache: bool = False):
        """
        Return the record for the given doi.

        Parameters
        ----------
        doi : str
            The DOI to be retrieved.
        errors : str
            Whether to ignore or raise errors.
        force : bool
            Whether to force the cache to retrieve a new entry.
        ignore_cache : bool
            Whether to use or ignore the cache.

        Returns
        -------
        record : requests.Response
            The response from Unpaywall.
        """
        record = None

        if not ignore_cache:
            if (doi not in self.content) or self.timed_out(doi) or force:
                downloaded = self.download(doi, errors)
                if downloaded:
                    self.access_times[doi] = time.time()
                    self.content[doi] = downloaded
                    self.save()
                    record = downloaded
            else:
                record = deepcopy(self.content[doi])
        else:
            record = self.download(doi, errors)
        return record

    def save(self, name=None) -> None:
        """
        Save the current cache contents to a file.

        Parameters
        ----------
        name : str or None
            The filename that the cache will be saved to.
            If None, self.name will be used.
        """
        if not name:
            name = self.name
        with open(name, 'wb') as handle:
            pickle.dump({'content': self.content,
                         'access_times': self.access_times},
                        handle)

    def load(self, name=None) -> None:
        """
        Load the cache from a file.

        Parameters
        ----------
        name : str or None
            The filename that the cache will be loaded from.
            If None, self.name will be used.
        """
        if not name:
            name = self.name
        with open(name, 'rb') as handle:
            data = pickle.load(handle)
        self.content = data['content']
        self.access_times = data['access_times']

    def download(self, doi: str, errors: str):
        """
        Retrieve a record from Unpaywall.

        Parameters
        ----------
        doi : str
            The DOI to be retrieved.
        errors : str
            Whether to ignore or raise errors.
        """
        from .utils import UnpywallURL

        mandatory_wait_time = int(os.environ.get('MANDATORY_WAIT_TIME', 1))
        time.sleep(mandatory_wait_time)
        url = UnpywallURL(doi=doi).doi_url

        try:

            r = requests.get(url)
            r.raise_for_status()
            return r

        # if DOI is invalid
        except requests.exceptions.HTTPError as HTTPError:
            if errors == 'raise':
                raise HTTPError

        except requests.exceptions.RequestException as RequestException:
            if errors == 'raise':
                raise RequestException

        # if bad internet connection
        except requests.exceptions.ConnectionError as ConnectionError:
            if errors == 'raise':
                raise ConnectionError

        # server is down
        except requests.exceptions.Timeout as Timeout:
            if errors == 'raise':
                raise Timeout

        warnings.warn('Could not download doi: {}'.format(doi))



import pytest

from unpywall.utils import UnpywallCredentials, UnpywallURL


class TestUnpywallCredentials:

    def test_validate_email(self):

        correct_email = 'nick.haupka@gmail.de'
        bad_email = 'https://github.com/naustica/unpywall'

        with pytest.raises(ValueError,
                           match=('An email address is required in order'
                                  ' to work with the Unpaywall API')):
            assert UnpywallCredentials.validate_email(None)

        with pytest.raises(ValueError,
                           match=('No valid email address entered.'
                                  ' Enter a valid email address')):
            assert UnpywallCredentials.validate_email(bad_email)

        with pytest.raises(ValueError, match='Do not use example.com'):
            assert UnpywallCredentials.validate_email(
              'nick.haupka@example.com')

        assert UnpywallCredentials.validate_email(
            correct_email) == correct_email

        assert isinstance(repr(UnpywallCredentials(correct_email)), str)

        assert isinstance(UnpywallCredentials(correct_email).email, str)


class TestUnpywallURL:

    def test_doi_url(self):

        doi = '10.1038/nature12373'

        url = UnpywallURL(doi=doi).doi_url

        assert isinstance(url, str)

    def test_query_url(self):

        query = 'duck'

        url = UnpywallURL(query=query).query_url

        assert isinstance(url, str)



import pytest
import os
import time
import uuid
from requests import Response
from shutil import copyfile

from unpywall.cache import UnpywallCache

os.environ['UNPAYWALL_EMAIL'] = 'bganglia892@gmail.com'


class TestUnpywallCache:

    test_dir = os.path.abspath(os.path.dirname(__file__))
    test_backup_cache_path = os.path.join(test_dir, 'unpaywall_cache')
    test_cache_path = os.path.join(test_dir, 'test_cache')

    @pytest.fixture
    def backup_cache(self):
        cache = UnpywallCache(TestUnpywallCache.test_backup_cache_path)
        yield cache

    @pytest.fixture
    def example_cache(self):

        if os.path.exists(TestUnpywallCache.test_cache_path):
            os.remove(TestUnpywallCache.test_cache_path)
            assert UnpywallCache(
                    TestUnpywallCache.test_cache_path).content == {}
            assert UnpywallCache(
                    TestUnpywallCache.test_cache_path).access_times == {}

        copyfile(TestUnpywallCache.test_backup_cache_path,
                 TestUnpywallCache.test_cache_path)
        cache = UnpywallCache(TestUnpywallCache.test_cache_path)
        assert cache.content != {}
        assert cache.access_times != {}
        yield cache

    def test_reset_cache(self, example_cache):
        example_cache.reset_cache()
        assert example_cache.content == {}
        assert example_cache.access_times == {}

    def test_delete(self, example_cache):
        doi = '10.1016/j.tmaid.2020.101663'
        assert doi in example_cache.content
        assert doi in example_cache.access_times
        example_cache.delete(doi)
        assert doi not in example_cache.content
        assert doi not in example_cache.access_times

    def test_timeout(self, example_cache):
        timeout = 1
        doi = '10.1016/j.jns.2020.116832'
        example_cache.timeout = timeout
        example_cache.delete(doi)
        assert doi not in example_cache.content
        example_cache.get(doi)
        time.sleep(timeout)
        assert example_cache.timed_out(doi)

    def test_get(self, example_cache, backup_cache):
        doi = '10.1016/j.jns.2020.116832'
        example_cache.delete(doi)
        assert doi not in example_cache.content
        assert doi not in example_cache.access_times
        assert isinstance(backup_cache.get(doi), Response)
        assert isinstance(example_cache.get(doi, ignore_cache=True), Response)

    def test_save_load(self, example_cache, backup_cache):
        doi = '10.1016/j.jns.2020.116832'
        example_cache.delete(doi)
        assert doi not in example_cache.content
        assert doi not in example_cache.access_times
        backup_cache.get(doi)
        saved_cache_name = str(uuid.uuid4().hex)
        example_cache.save(saved_cache_name)
        UnpywallCache(saved_cache_name)
        os.remove(saved_cache_name)
        path = os.path.join(os.getcwd(), 'unpaywall_cache')
        assert UnpywallCache().name == path
        os.remove(path)
        assert doi in backup_cache.content
        assert doi in backup_cache.access_times

    def test_download(self, example_cache):
        doi = '10.1016/j.jns.2020.116832'
        assert isinstance(example_cache.download(doi, 'raise'), Response)



import requests
import pandas as pd
import sys
import subprocess
import tempfile
import webbrowser
import os
import platform
from io import BytesIO
from functools import reduce

from .cache import UnpywallCache


class Unpywall:
    """
    Base class that contains useful functions for retrieving information
    from the Unpaywall REST API (https://api.unpaywall.org). This client uses
    version 2 of the API.
    """

    api_limit: int = 100000
    cache = None

    @staticmethod
    def init_cache(cache=None) -> None:
        """
        This method initilializes a cache that is used to store records
        from the Unpaywall database.

        Parameters
        ----------
        cache: UnpywallCache
            A custom cache to be used instead of the standard cache.

        Raises
        ------
        AttributeError
            If the custom cache is not of type UnpywallCache.
        """
        if cache:
            if not isinstance(cache, UnpywallCache):
                raise AttributeError(
                 'Cache is not of type {0}'.format(UnpywallCache))
            else:
                Unpywall.cache = cache
        else:
            Unpywall.cache = UnpywallCache()

    @staticmethod
    def _validate_dois(dois: list) -> list:
        """
        This method accepts a list of DOIs and returns a cleaned version of it.
        Raises an error if the desired input is not given.

        Parameters
        ----------
        dois : list
            A list of DOIs. Each DOI should be represented as a string.

        Returns
        -------
        list
            A list of DOIs.

        Raises
        ------
        ValueError
            If the input is empty or not of type list.
        """
        if dois is None or len(dois) == 0:
            raise ValueError('No DOI specified')

        if not isinstance(dois, list):
            raise ValueError('The input format must be of type list')

        if len(dois) > Unpywall.api_limit:
            raise ValueError('Unpaywall only allows to 100,000 calls per day')

        for doi in dois:
            doi.replace(' ', '')

        return dois

    @staticmethod
    def _progress(progress: float) -> None:
        """
        This method prints out the current progress status of an API call.

        Parameters
        ----------
        progress : float
            Current status of the progress. Value between 0 and 1.
        """

        bar_len = 50
        block = int(round(bar_len * progress))

        text = '|{0}| {1}%'.format('=' * block + ' ' * (bar_len - block),
                                   int(progress * 100))

        print(text, end='\r', flush=False, file=sys.stdout)

        if progress == 1:
            print('\n', file=sys.stdout)

    @staticmethod
    def _get_df(data,
                format: str,
                errors: str) -> pd.DataFrame:
        """
        Parses information from the Unpaywall API service and returns it as
        a pandas DataFrame.

        Parameters
        ----------
        data: JSON object
            A JSON data structure containing all information
            returned by Unpaywall about a given input.
        format: str
            The format of the DataFrame.
        errors : str
            Either 'raise' or 'ignore'. If the parameter errors is set to
            'ignore' than errors will not raise an exception.

        Returns
        -------
        DataFrame
            A pandas DataFrame that contains information from the Unpaywall
            API service.

        Raises
        ------
        ValueError
            If the parameter errors contains a faulty value.
        """

        if format not in ['raw', 'extended']:
            raise ValueError('The argument format only accepts the'
                             ' values "raw" and "extended"')

        if format == 'extended':

            doi_object = pd.json_normalize(data=data,
                                           max_level=1,
                                           errors=errors)

            doi_object.drop(columns=['oa_locations',
                                     'z_authors'],
                            errors=errors,
                            inplace=True)

            oa_locations = pd.json_normalize(data=data,
                                             errors=errors,
                                             meta='doi',
                                             record_path=['oa_locations'])

            z_authors = pd.json_normalize(data=data,
                                          errors=errors,
                                          meta='doi',
                                          record_path=['z_authors'])

            dfs = [doi_object, oa_locations, z_authors]
            df = reduce(lambda left, right: pd.merge(left,
                                                     right,
                                                     how='outer',
                                                     on='doi'), dfs)

        else:
            df = pd.json_normalize(data=data, max_level=1, errors=errors)

        return df

    @staticmethod
    def query(query: str,
              is_oa: bool = False,
              format: str = 'raw',
              errors: str = 'raise') -> pd.DataFrame:
        """
        Parses information for a given query from the Unpaywall API service and
        returns it as a pandas DataFrame.

        Parameters
        ----------
        query : str
            The text to search for.
        is_oa : bool
            A boolean value indicating whether the returned records should be
            Open Access or not.
        format: str
            The format of the DataFrame.
        errors : str
            Either 'raise' or 'ignore'. If the parameter errors is set to
            'ignore' than errors will not raise an exception.

        Returns
        -------
        DataFrame
            A pandas DataFrame that contains information from the Unpaywall
            API service.
        """

        data = Unpywall.get_json(query=query, is_oa=is_oa, errors=errors)

        df = pd.DataFrame()

        for obj in data['results']:
            df2 = Unpywall._get_df(data=obj['response'],
                                   format=format,
                                   errors=errors)
            df = pd.concat([df, df2], ignore_index=True)

        if df.empty:
            return None

        return df

    @staticmethod
    def doi(dois: list,
            format: str = 'raw',
            progress: bool = False,
            errors: str = 'raise',
            force: bool = False,
            ignore_cache: bool = False):
        """
        Parses information for a given DOI from the Unpaywall API service and
        returns it as a pandas DataFrame.

        Parameters
        ----------
        dois : list
            A list of DOIs.
        format: str
            The format of the DataFrame.
        progress : bool
            Whether the progress of the API call should be printed out or not.
        errors : str
            Either 'raise' or 'ignore'. If the parameter errors is set to
            'ignore' than errors will not raise an exception.
        force : bool
            Whether to force the cache to retrieve a new entry.
        ignore_cache : bool
            Whether to use or ignore the cache.

        Returns
        -------
        DataFrame
            A pandas DataFrame that contains information from the Unpaywall
            API service.
        """

        dois = Unpywall._validate_dois(dois)

        df = pd.DataFrame()

        for n, doi in enumerate(dois, start=1):

            if progress:
                Unpywall._progress(n / len(dois))

            data = Unpywall.get_json(doi,
                                     errors=errors,
                                     force=force,
                                     ignore_cache=ignore_cache)

            # check if json is not empty or None due to an faulty DOI
            if not bool(data):
                continue

            df2 = Unpywall._get_df(data=data,
                                   format=format,
                                   errors=errors)
            df = pd.concat([df, df2], ignore_index=True)

        if df.empty:
            return None

        return df

    @staticmethod
    def get_json(doi: str = None,
                 query: str = None,
                 is_oa: bool = False,
                 errors: str = 'raise',
                 force: bool = False,
                 ignore_cache: bool = False):
        """
        This function returns all information in Unpaywall about the given DOI.

        Parameters
        ----------
        doi : str
            The DOI of the requested paper.
        query : str
            The text to search for.
        is_oa : bool
            A boolean value indicating whether the returned records should be
            Open Access or not.
        errors : str
            Either 'raise' or 'ignore'. If the parameter errors is set to
            'ignore' than errors will not raise an exception.
        force : bool
            Whether to force the cache to retrieve a new entry.
        ignore_cache : bool
            Whether to use or ignore the cache.

        Returns
        -------
        JSON object
            A JSON data structure containing all information
            returned by Unpaywall about the given DOI.

        Raises
        ------
        AttributeError
            If the Unpaywall API did not respond with json.
        """
        if not Unpywall.cache:
            Unpywall.init_cache()

        if doi:
            r = Unpywall.cache.get(doi,
                                   errors=errors,
                                   force=force,
                                   ignore_cache=ignore_cache)
        if query:

            if type(is_oa) != bool:
                raise ValueError('The argument is_oa only accepts the'
                                 ' values "True" and "False"')
            # TODO: implementation -> cache.py
            from .utils import UnpywallURL

            url = UnpywallURL(query=query, is_oa=is_oa).query_url

            r = requests.get(url)
        try:
            return r.json()
        except AttributeError:
            return None

    @staticmethod
    def get_pdf_link(doi: str) -> str:
        """
        This function returns a link to an OA pdf (if available).

        Parameters
        ----------
        doi: str
            The DOI of the requested paper.

        Returns
        -------
        str
            The URL of an OA PDF (if available).
        """
        json_data = Unpywall.get_json(doi)
        try:
            return json_data['best_oa_location']['url_for_pdf']
        except (KeyError, TypeError):
            return None

    @staticmethod
    def get_doc_link(doi: str) -> str:
        """
        This function returns a link to the best OA location
        (not necessarily a PDF).

        Parameters
        ----------
        doi : str
            The DOI of the requested paper.

        Returns
        -------
        str
            The URL of the best OA location (not necessarily a PDF).
        """
        json_data = Unpywall.get_json(doi)
        try:
            return json_data['best_oa_location']['url']
        except (KeyError, TypeError):
            return None

    @staticmethod
    def get_all_links(doi: str) -> list:
        """
        This function returns a list of URLs for all open-access copies
        listed in Unpaywall.

        Parameters
        ----------
        doi : str
            The DOI of the requested paper.

        Returns
        -------
        list
            A list of URLs leading to open-access copies.
        """
        data = []
        for value in [Unpywall.get_doc_link(doi),
                      Unpywall.get_pdf_link(doi)]:
            if value and value not in data:
                data.append(value)
        return data

    @staticmethod
    def download_pdf_handle(doi: str) -> BytesIO:
        """
        This function returns a file-like object containing the requested PDF.

        Parameters
        ----------
        doi : str
            The DOI of the requested paper.

        Returns
        -------
        BytesIO
            The handle of the PDF file.
        """
        pdf_link = Unpywall.get_pdf_link(doi)
        r = requests.get(pdf_link)
        return BytesIO(bytearray(r.text, encoding='utf-8'))

    @staticmethod
    def view_pdf(doi: str,
                 mode: str = 'viewer',
                 progress: bool = False) -> None:
        """
        This function opens a local copy of a PDF from a given DOI.

        Parameters
        ----------
        doi : str
            The DOI of the requested paper.
        mode : str
            The mode for viewing a PDF.
        progress : bool
            Whether the progress of the API call should be printed out or not.
        """

        url = Unpywall.get_pdf_link(doi)
        r = requests.get(url, stream=url)
        file_size = int(r.headers.get('content-length', 0))
        block_size = 1024

        if mode == 'viewer':

            tmp = tempfile.NamedTemporaryFile(delete=False, suffix='.pdf')

            with open(tmp.name, 'wb') as file:
                chunk_size = 0
                for chunk in r.iter_content(block_size):
                    if progress:
                        chunk_size += len(chunk)
                        Unpywall._progress(chunk_size / file_size)
                    file.write(chunk)

                # macOS
                if platform.system() == 'Darwin':
                    subprocess.run(['open', tmp.name], check=True)
                # Windows
                elif platform.system() == 'Windows':
                    os.startfile(tmp.name)
                # Linux
                else:
                    subprocess.run(['xdg-open', tmp.name], check=True)

        else:
            webbrowser.open_new(url)

    @staticmethod
    def download_pdf_file(doi: str,
                          filename: str,
                          filepath: str = '.',
                          progress: bool = False) -> None:
        """
        This function downloads a PDF from a given DOI.

        Parameters
        ----------
        doi : str
            The DOI of the requested paper.
        filename : str
            The filename for the PDF.
        filepath : str
            The path to store the downloaded PDF.
        progress : bool
            Whether the progress of the API call should be printed out or not.
        """

        url = Unpywall.get_pdf_link(doi)
        r = requests.get(url, stream=url)
        file_size = int(r.headers.get('content-length', 0))
        block_size = 1024

        path = os.path.join(filepath, filename)

        if not os.path.exists(filepath):
            os.makedirs(filepath)

        with open(path, 'wb') as file:
            chunk_size = 0
            for chunk in r.iter_content(block_size):
                if progress and file_size > 0:
                    chunk_size += len(chunk)
                    Unpywall._progress(chunk_size / file_size)
                file.write(chunk)



import pytest
import pandas as pd
import os
from io import BytesIO
from requests.exceptions import HTTPError

from unpywall import Unpywall
from unpywall.cache import UnpywallCache

test_cache = UnpywallCache(os.path.join(
    os.path.abspath(
       os.path.dirname(__file__)), 'unpaywall_cache'))

os.environ['UNPAYWALL_EMAIL'] = 'nick.haupka@gmail.com'


class TestUnpywall:

    @pytest.fixture
    def Unpywall(self):
        Unpywall.init_cache(test_cache)
        yield Unpywall

    def test_init_cache(self):

        Unpywall.init_cache()
        path = os.path.join(os.getcwd(), 'unpaywall_cache')
        assert Unpywall.cache.name == path
        os.remove(path)

        with pytest.raises(
         AttributeError, match='Cache is not of type {0}'.format(
                UnpywallCache)):
            assert Unpywall.init_cache('Not a UnpywallCache object.')

        Unpywall.init_cache(test_cache)
        assert os.path.exists(Unpywall.cache.name)

    def test_validate_dois(self):

        correct_dois = ['10.1038/nature12373', '10.1103/physreve.88.012814']
        bad_dois = 'a bad doi'

        with pytest.raises(ValueError, match='No DOI specified'):
            assert Unpywall._validate_dois(None)

        with pytest.raises(ValueError,
                           match='The input format must be of type list'):
            assert Unpywall._validate_dois(bad_dois)

        with pytest.raises(ValueError,
                           match=('Unpaywall only allows to 100,000 calls'
                                  ' per day')):
            assert Unpywall._validate_dois(['doi'] * (Unpywall.api_limit + 1))

        assert Unpywall._validate_dois(correct_dois) == correct_dois

    def test_get_df(self, Unpywall):

        data = Unpywall.get_json(doi='10.1016/j.tmaid.2020.101663',
                                 errors='raise')

        with pytest.raises(ValueError,
                           match=('The argument format only accepts the'
                                  ' values "raw" and "extended"')):
            assert Unpywall._get_df(data=data,
                                    format='not a valid format',
                                    errors='raise')

        df_raw = Unpywall._get_df(data=data,
                                  format='raw',
                                  errors='ignore')

        assert isinstance(df_raw, pd.DataFrame)

        df_extended = Unpywall._get_df(data=data,
                                       format='extended',
                                       errors='ignore')

        assert isinstance(df_extended, pd.DataFrame)

    def test_doi(self, Unpywall, capfd):

        df = Unpywall.doi(dois=['10.1038/nature12373'],
                          format='raw',
                          progress=True,
                          errors='ignore')

        captured = capfd.readouterr()

        assert len(captured.out) > 0
        assert isinstance(df, pd.DataFrame)

        with pytest.warns(UserWarning):
            df_empty = Unpywall.doi(dois=['a bad doi'],
                                    errors='ignore')

            assert df_empty is None

    def test_query(self, Unpywall):

        df = Unpywall.query(query='test',
                            is_oa=True,
                            errors='ignore')

        assert isinstance(df, pd.DataFrame)

    def test_get_json(self, Unpywall):

        with pytest.raises(HTTPError):
            Unpywall.get_json(doi='a bad doi', errors='raise')

        with pytest.warns(UserWarning):
            Unpywall.get_json(doi='a bad doi', errors='ignore')

        assert isinstance(Unpywall.get_json(doi='10.1016/j.tmaid.2020.101663',
                                            errors='raise'), dict)

        assert isinstance(Unpywall.get_json(query='test',
                                            is_oa=True,
                                            errors='raise'), dict)

        with pytest.raises(ValueError,
                           match=('The argument is_oa only accepts the'
                                  ' values "True" and "False"')):
            assert Unpywall.get_json(query='test',
                                     is_oa='test',
                                     errors='raise')

    def test_get_pdf_link(self, Unpywall):

        assert isinstance(Unpywall.get_pdf_link('10.1038/nature12373'), str)

    def test_get_doc_link(self, Unpywall):

        assert isinstance(
            Unpywall.get_doc_link('10.1016/j.tmaid.2020.101663'), str)

    def test_get_all_links(self, Unpywall):

        assert isinstance(
            Unpywall.get_all_links('10.1016/j.tmaid.2020.101663'), list)

    def test_download_pdf_handle(self, Unpywall):

        assert isinstance(
            Unpywall.download_pdf_handle('10.1038/nature12373'), BytesIO)

    def test_progress(self, Unpywall, capfd):
        Unpywall._progress(0.5)
        captured = capfd.readouterr()
        assert len(captured.out) > 0

    def test_view_pdf(self, Unpywall):
        pass

    def test_download_pdf_file(self, Unpywall, capfd):

        filename = 'test.pdf'
        filepath = './test_dir'

        Unpywall.download_pdf_file('10.1038/nature12373',
                                   filename=filename,
                                   filepath=filepath,
                                   progress=True)

        captured = capfd.readouterr()
        assert len(captured.out) > 0

        path = os.path.join(filepath, filename)
        assert os.path.exists(path)

        os.remove(path)
        os.rmdir(filepath)



from unpywall.__main__ import main
from requests.exceptions import HTTPError
import pytest
import os

os.environ['UNPAYWALL_EMAIL'] = 'nick.haupka@gmail.com'


class TestUnpywallCli:

    def test_main(self):

        with pytest.raises(SystemExit) as pytest_raise_system_exit:
            assert main(test_args=['-h'])

            assert pytest_raise_system_exit.value.code == 0

        with pytest.raises(SystemExit) as pytest_raise_system_exit:
            main(test_args=['this is a bad argument'])
            assert pytest_raise_system_exit.value.code == 1

    def test_view(self):
        with pytest.raises(SystemExit) as pytest_raise_system_exit:
            main(test_args=['view'])

            assert pytest_raise_system_exit.value.code == 2

        with pytest.raises(HTTPError):
            bad_doi = 'bad_doi'
            main(test_args=['view', bad_doi])

    def test_download(self, capfd):

        bad_doi = 'bad_doi'
        main(test_args=['download', bad_doi])
        captured = capfd.readouterr()
        assert captured.out == 'Could not download file.\n'

        with pytest.raises(SystemExit) as pytest_raise_system_exit:
            main(test_args=['download'])

            assert pytest_raise_system_exit.value.code == 2

    def test_link(self):
        with pytest.raises(SystemExit) as pytest_raise_system_exit:
            main(test_args=['link'])

            assert pytest_raise_system_exit.value.code == 2

        with pytest.raises(HTTPError):
            bad_doi = 'bad_doi'
            main(test_args=['link', bad_doi])



from setuptools import setup
from os import path

dir = path.abspath(path.dirname(__file__))
with open(path.join(dir, 'README.md'), encoding='utf-8') as file:
    long_description = file.read()


setup(name='unpywall',
      version='0.2.1',
      description='Interfacing the Unpaywall Database with Python',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='https://github.com/unpywall/unpywall',
      author='Nick Haupka, bganglia',
      author_email='nick.haupka@gmail.com, bganglia892@gmail.com',
      license='MIT',
      packages=['unpywall'],
      keywords=['Unpaywall', 'Open Access', 'full text'],
      project_urls={
        'Documentation': 'https://unpywall.readthedocs.io/en/latest/',
        'Source': 'https://github.com/unpywall/unpywall',
        'Tracker': 'https://github.com/unpywall/unpywall/issues'
      },
      install_requires=[
        'pandas>=2.0',
        'requests'
      ],
      extras_require={
       'dev': [
           'pytest',
           'coverage',
           'pytest-cov',
           'sphinx',
           'alabaster'
       ]
      },
      entry_points={
        'console_scripts': [
            'unpywall = unpywall.__main__:main'
        ]
      },
      classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9'
      ],
      zip_safe=False)



import os
import re


class UnpywallCredentials:
    """
    This class provides tools for setting up an email for the
    Unpaywall service.

    Attributes
    ----------
    email : str
        An email that is necessary for using the Unpaywall API service.
    """

    def __init__(self, email: str) -> None:
        self.email = email

    def __repr__(self) -> str:
        return 'Your email has been set.'

    @property
    def email(self) -> str:
        return self._email

    @email.setter
    def email(self, user_email: str) -> None:

        user_email = UnpywallCredentials.validate_email(user_email)

        os.environ['UNPAYWALL_EMAIL'] = user_email

        self._email = user_email

    @staticmethod
    def validate_email(email: str) -> str:
        """
        This method takes an email as input and raises an error if the email is
        not valid. Otherwise the email will be returned.

        Parameters
        ----------
        email : str
            An email that is necessary for using the Unpaywall API service.

        Returns
        -------
        str
            The email that was given as input.

        Raises
        ------
        ValueError
            If the email parameter is empty or not valid.
        """

        # from https://stackoverflow.com/a/43937713/12580727
        email_regex = r'^[\w\.\+\-]+\@[\w]+\.[a-z]{2,3}$'

        if email is None:
            raise ValueError(('An email address is required in order'
                              ' to work with the Unpaywall API'))

        if not re.match(email_regex, email):
            raise ValueError(('No valid email address entered. Enter a'
                              ' valid email address'))

        if 'example.com' in email:
            raise ValueError('Do not use example.com')

        return email


class UnpywallURL:
    """
    This class provides the Unpaywall URL.

    Attributes
    ----------
    doi : str
        The DOI of the requested paper.
    query : str
        The text to search for.
    is_oa : bool
        A boolean value indicating whether the returned records should be
        Open Access or not.
    doi_url : str
        The URL for the DOI-Endpoint.
    query_url : str
        The URL for the Query-Endpoint
    """

    def __init__(self,
                 doi: str = None,
                 query: str = None,
                 is_oa: bool = False) -> None:

        self.doi = doi
        self.query = query
        self.is_oa = is_oa
        self.email = UnpywallCredentials.validate_email(
            os.environ.get('UNPAYWALL_EMAIL'))

    @property
    def doi_url(self) -> str:

        if self.doi is None:
            raise ValueError('Missing DOI')

        return 'https://api.unpaywall.org/v2/{0}?email={1}'.format(self.doi,
                                                                   self.email)

    @property
    def query_url(self) -> str:

        if self.query is None:
            raise ValueError('Missing query')
        return ('https://api.unpaywall.org/v2/search/?query={0}&is_oa={1}'
                '&email={2}').format(self.query, self.is_oa, self.email)



from argparse import ArgumentParser, RawTextHelpFormatter, SUPPRESS
import textwrap
import uuid
import sys

from unpywall import Unpywall


class UnpywallArgumentParser(ArgumentParser):

    def error(self, message: str) -> None:
        print('error: {0}\n'.format(message), end='\n', file=sys.stderr)
        self.print_help()
        sys.exit(2)


class main:

    def __init__(self, test_args=None) -> None:
        self.test_args = test_args
        usage = textwrap.dedent("""unpywall <command> [<args>]

                \nCommand-line tool for interfacing the Unpaywall API
                                """)

        description = textwrap.dedent("""
                 These are common unpywall commands:

                    view        This command opens a local copy of a PDF from
                                a given DOI.
                    download    This command downloads a copy of a PDF from a
                                given DOI.
                    link        This command returns a link to an OA pdf
                                (if available).
                                      """)
        ap = UnpywallArgumentParser(prog='unpywall',
                                    usage=usage,
                                    description=description,
                                    formatter_class=RawTextHelpFormatter,
                                    add_help=False)

        ap.add_argument('command', help=SUPPRESS)
        ap.add_argument('-h',
                        '--help',
                        action='help',
                        default=SUPPRESS,
                        help=SUPPRESS)
        if self.test_args:
            args = ap.parse_args(self.test_args[0:1])
        else:
            args = ap.parse_args(sys.argv[1:2])
        if not hasattr(self, args.command):
            print('Unknown option: {}'.format(args.command))
            ap.print_help()
            sys.exit(1)
        getattr(self, args.command)()

    def __repr__(self) -> None:
        return None

    def view(self) -> None:
        ap = UnpywallArgumentParser(description=('This command opens a local'
                                                 ' copy of a PDF from a'
                                                 ' given DOI.'),
                                    formatter_class=RawTextHelpFormatter,
                                    add_help=False)
        ap.add_argument('doi',
                        type=str,
                        metavar='doi',
                        help='\tThe DOI of the document.')
        ap.add_argument('-m',
                        '--mode',
                        type=str,
                        default='viewer',
                        dest='mode',
                        choices=['viewer', 'browser'],
                        metavar='\b',
                        help='\tThe mode for viewing a PDF.')
        ap.add_argument('-b',
                        '--backend',
                        type=str,
                        default='remote',
                        dest='backend',
                        choices=['remote', 'cache', 'snapshot'],
                        metavar='\b',
                        help='\tThe backend you want to use.')
        ap.add_argument('-u',
                        '--progress',
                        type=bool,
                        default=False,
                        dest='progress',
                        metavar='\b',
                        help='\tShow progress bar.')
        ap.add_argument('-h',
                        '--help',
                        action='help',
                        default=SUPPRESS,
                        help=SUPPRESS)

        if self.test_args:
            args = ap.parse_args(self.test_args[1:])
        else:
            args = ap.parse_args(sys.argv[2:])

        Unpywall.view_pdf(args.doi, args.mode, progress=args.progress)

    def download(self) -> None:
        ap = UnpywallArgumentParser(description=('This command downloads a'
                                                 ' copy of a PDF from a'
                                                 ' given DOI.'),
                                    formatter_class=RawTextHelpFormatter,
                                    add_help=False)
        ap.add_argument('doi',
                        type=str,
                        metavar='doi',
                        help='\tThe DOI of the document.')
        ap.add_argument('-f',
                        '--filename',
                        type=str,
                        dest='filename',
                        metavar='\b',
                        help='\tThe filename for downloading a PDF.')
        ap.add_argument('-p',
                        '--path',
                        type=str,
                        default='.',
                        dest='filepath',
                        metavar='\b',
                        help='\tThe filepath for downloading a PDF.')
        ap.add_argument('-b',
                        '--backend',
                        type=str,
                        default='remote',
                        dest='backend',
                        choices=['remote', 'cache', 'snapshot'],
                        metavar='\b',
                        help='\tThe backend you want to use.')
        ap.add_argument('-u',
                        '--progress',
                        type=bool,
                        default=False,
                        dest='progress',
                        metavar='\b',
                        help='\tShow progress bar.')
        ap.add_argument('-h',
                        '--help',
                        action='help',
                        default=SUPPRESS,
                        help=SUPPRESS)

        if self.test_args:
            args = ap.parse_args(self.test_args[1:])
        else:
            args = ap.parse_args(sys.argv[2:])

        try:
            if not args.filename:
                args.filename = '{0}.pdf'.format(uuid.uuid4().hex)
            Unpywall.download_pdf_file(args.doi,
                                       filename=args.filename,
                                       filepath=args.filepath,
                                       progress=args.progress)
            print('File was successfully downloaded.')
        except Exception:
            print('Could not download file.')

    def link(self) -> None:
        ap = UnpywallArgumentParser(description=('This command returns a link'
                                                 ' to an OA pdf'
                                                 ' (if available).'),
                                    formatter_class=RawTextHelpFormatter,
                                    add_help=False)
        ap.add_argument('doi',
                        type=str,
                        metavar='doi',
                        help='\tThe DOI of the document.')
        ap.add_argument('-b',
                        '--backend',
                        type=str,
                        default='remote',
                        dest='backend',
                        choices=['remote', 'cache', 'snapshot'],
                        metavar='\b',
                        help='\tThe backend you want to use.')
        ap.add_argument('-h',
                        '--help',
                        action='help',
                        default=SUPPRESS,
                        help=SUPPRESS)

        if self.test_args:
            args = ap.parse_args(self.test_args[1:])
        else:
            args = ap.parse_args(sys.argv[2:])

        print(Unpywall.get_pdf_link(args.doi))


if __name__ == '__main__':
    main()
